///////////////////////////////////////////////////////////////////////////////
//
// midi-faders.js
//
// Mapping logic for 16 midi faders.  We conceive of 'banks' of faders, each of
// which has 16 controllers and is dedicated to a particular task e.g. loop
// control, granulation etc.  We can switch banks by sending a changeBank
// message to the left inlet (handled automatically by the umenu object in the
// interface patch).  Fader values are sent to the right (channel) and middle
// (midi value) inlets (create a simple subpatch/abstraction to create this
// e.g. ld2-midi-controller2).
//
// There can be as many banks as we wish; they are numbered integers starting
// from 1.  A bank consists of 16 controller objects which specify, scale, and
// hold the data for a midi-fader (e.g. last value, min/max/exponent for
// scaling and curves, label (for the interface), and send name (i.e. the
// user-specified global messenger that transmits the scaled data).  
//
// We can switch banks safely in that a single fader's value is remembered from
// bank to bank, so when the bank is switched and a fader is moved, it creates
// no output until it's within a specified tolerance (default 3) of the
// previous value on this bank.  

// When the bank is changed, the labels under the fader numbers will change
// accordingly by receiving e.g. midi-faders-js-label1.  The numbers displayed
// are also changed via e.g. midi-faders-js-number1.  NB when you change banks
// and a fader is outside the tolerance, the display will show the target MIDI
// value in red before outputting again.
//
// Mapping is done in a text file which coll reads (see
// midi-faders-js-banks.txt for an example).  The indices of these colls need
// not be unique as we won't use it for lookup, rather just for dumping one
// line at a time.  The format of each coll line is:
// 
// bank number, channel out-min out-max exponent initial-value receiver label;
// e.g. 1, 8 -13 13 0.9 0 1-sndfiles-speed-st sf1-st;
//
// If initial-value is outside the range of out-min and out-max, it will be
// ignored, otherwise it will be used as the fader target value when it's first
// moved. 
//
// Any controller channels can be designated 'universal'.  In this case they
// will be active no matter which bank is active; as such the tolerance
// mechanism is also dispensed with.  Such controllers are indicated on bank 0.
// It would be an error to try and redefine a controller channel to a specific
// bank when it has already been designated as universal.
// 
// Further, any controller can be a link to another, so as to have
// exactly the same functionality as the one it points too, but without being
// universal.  This is indicated by giving the usual bank and channel, but
// then instead of out-min and out-max, the bank and channel of the fader we
// want to link to, an exponent of -1, and nothing for initial-value, receiver,
// or label e.g.  
// 3, 11 1 8 -1;
// This means that fader 11 on bank 3 merely links to fader 8 of bank 1.
//
// The scaling is done to any arbitrary floating-point range so MIDI 0 will
// become 'out-min' and MIDI 127 will be 'out-max'.  An exponent is given to
// create a curve: 1 would be a straight line, > 1.0 would be concave (sagging)
// and < 1.0 would be convex.
//
// There's functionality to change the scaling data or force current values
// etc.  See the global functions below for specifics.
//
// Todo:
// allow one fader to control many outputs (one-to-many) by having an array of
// controller objects in the bank instead of just a single instance.
//
// history:
// 23.9.13: added an argument to toggleOnOFF and setToggle function to set
// toggles explicitly to 1 or 1
// 
// 6.10.11: added toggleOnOff function to be able to temporarily output a 0 but
// then return to previous value.  This is done by sending a channel number to
// the 4th inlet; it will act on the currently selected bank and will follow
// links. 
// 
// SVN ID: $Id: midi-faders.js 1248 2010-09-04 15:09:23Z medward2 $
//
// * $$ Last modified: 14:34:00 Mon Sep 23 2013 BST
//
// Michael Edwards ~ m@michael-edwards.org ~ 13.7.10 (Ayuthaya)
//
///////////////////////////////////////////////////////////////////////////////
//
// This is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
// 
// This software is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// this software; if not, write to the Free Software Foundation, Inc., 59
// Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
////////////////////////////////////////////////////////////////////////////////

// inlet data:
// 1) bank index (integer) 
// 2) midi-fader value; this should always be preceeded by the channel in inlet
//    3 
// 3) midi-fader channel (counting channels from 1-16)
// 4) midi-fader channel to toggle on/off state for
// 5) a pedal (or some such other 7bit) value used to scale one of the
//    controllers. 
inlets = 5;
// despite the fact we send out scaled data via sends, we'll send it out the
// outlet too.
outlets = 1;

// global configuration data; hardly a need for a class here but I usually need
// much more data so usually do it this way to preserve namespace.
var MidiFadersConfig = {
  // minimum 7-bit distance to the previous value before starting to output.
  tolerance : 1,
  // set this to false if you don't want to see warnings when an unmapped
  // controller is used etc. 
  warn : true,
}

// our main midi faders banks object to do all the work.
var gMidiFadersBanks = new MidiFadersBanks();

// a global to allow us to send messages to receive objects in the max patches.
var MidiFadersGlobalMessenger = new Global("MidiFadersGlobalMessenger");

////////////////////////////////////////////////////////////////////////////////
// Constructor
function MidiFadersBanks() {
  // so 100 banks (should be enough).  Each bank has 16 controllers.
  // bank 0 is the universal bank.
  this.banks = new Array(100); 
  this.currentBank = 1; // don't bother with slot 0 just yet
  this.currentChannel = -1;
  this.didStore = false;
  this.pedal = false; // controller object
  this.pedalScaledVal = 1.0;
  this.pedalController = false;
  this.pedalBank = -1;
  this.pedalChannel = -1;

  this.storeController = function(controller, bankNum, channel) {
    // don't allow storage of channels which have already been designated
    // universal 
    if (bankNum && this.universal(channel)) {
      post("\nmidi-faders.js::channel " + channel + " has already been " +
           "designated universal; ignoring (check your mapping text file).");
    }
    else {
      if (!this.banks[bankNum])
        this.banks[bankNum] = new bank(bankNum);
      this.banks[bankNum].storeController(controller, channel);
      this.didStore = true;
    }
  }
  this.checkStore = function() {
    if (this.didStore)
      return true;
    else {
      post("\nmidi-faders.js::Can't perform operation until mappings " +
           "file has been processed.");
      return false;
    }
  }
  //  this is the main function that does the scaling etc.
  this.output = function(val) { // val is 0-127
    // display the current fader value irrespective of what we're going to do
    // with it.  
    midiFadersSend("midi-faders-js-current-midi", val);
    if (this.checkStore()) {
      var controller = this.universal(this.currentChannel);
      if (!controller) // it's not universal
        controller = this.getController(this.currentBank, this.currentChannel,
                                        // could be we move an unmapped fader
                                        // so we don't want to warn
                                        false, true); // follow links
      // still could be we don't have a controller mapped to this channel...
      if (controller)
        // the controller checks if it's in a valid range or not
        controller.sendScaledVal(val);
    }
  }
  this.pedalIn = function(val) { // 0-127
    var scaled = 1.0;
    if (this.pedal) {
      scaled = this.pedal.scale(val);
      this.pedalScaledVal = scaled;
    }
    if (this.pedalController) {
      // post(this.pedalController.send);
      this.pedalController.rescale(scaled);
    }
  }
  this.getController = function(bank, channel, reallyWarn, getLink) {
    var bnk = false; 
    var ctlr = false;
    if (bank >= 0 && channel >= 1) {
      bnk = this.banks[bank];
      ctlr = bnk ? bnk.controllers[channel] : false;
    }
    if (ctlr) {
      if (ctlr.isLink && getLink)
        ctlr = this.getController(ctlr.linkBank, ctlr.linkChannel, reallyWarn,
                                  getLink);
      return ctlr;
    }
    else {
      if (MidiFadersConfig.warn && reallyWarn)
        post("\nmidi-faders.js::getController: can't get controller " + channel
             + " from bank " + bank);
      return false;
    }
  }
  this.toggleOnOff = function(channel) {
    // post("\ntoggleOnOff channel " + channel);
    ctlr = this.getController(this.currentBank, channel, false, true)
    if (ctlr)
      ctlr.toggleOnOff();
  }
  // MDE Mon Sep 23 14:28:37 2013 
  this.setToggle = function(channel, value) {
    ctlr = this.getController(this.currentBank, channel, false, true)
    if (ctlr)
      ctlr.toggleOnOff(value);
  }
  // returns the controller on the given channel, bank 0 (universal)
  this.universal = function(channel) {
    return this.getController(0, channel, false, false);
  }
  this.getControllerLabel = function(bank, channel) {
    var controller = this.getController(bank, channel, true, false);
    var label = controller ? controller.label : false;
    return label
  }
  this.changeBank = function(bankNum) {
    if (this.checkStore()) {
      this.currentBank = bankNum;
      // deactivate all banks
      this.deactivateAllBanks();
      bnk = this.banks[bankNum];
      var i;
      if (bnk) {
        // send the labels for this bank, overwriting all current labels. The
        // labels under each fader display on the interface change their text
        // by receiving e.g. midi-faders-js-label1
        bnk.sendLabels(true);
        // show the current controller values (scaled)
        bnk.displayCurrent();
      }
      // send the labels for the universal bank, without overwriting
      this.banks[0].sendLabels(false);
    }
  }
  // this sets the channel just before value() is called
  this.channel = function(chan) {
    // don't bother subtracting 1 to store from array index 0, just leave the 0
    // slot blank.
    this.currentChannel = chan;
  }
  this.print = function() {
    var i, j;
    for (i = 0; i < this.banks.length; ++i) {
      var bnk = this.banks[i];
      if (bnk) {
        for (j = 1; j < 17; ++j) {
          var ctlr = bnk.controllers[j];
          if (ctlr)
            ctlr.print();
        }
      }
    }
  }
  this.endStore = function() {
    this.handleLinks();
    changeBank(1);
  }
  this.deactivateAllBanks = function() {
    var i;
    for (i = 0; i < this.banks.length; ++i) {
      var bnk = this.banks[i];
      if (bnk) {
        bnk.deactivate();
      }
    }
  }
  // the main job here is 1) to make sure that a controller that is pointed to
  // from another actually exists 2) to copy over the label of the controller
  // pointed to
  this.handleLinks = function() {
    var i, j;
    for (i = 1; i < this.banks.length; ++i) {
      var bnk = this.banks[i];
      if (bnk) {
        for (j = 1; j < 17; ++j) {
          var ctlr = bnk.controllers[j];
          if (ctlr) {
            if (ctlr.isLink) {
              ctlr.label =
                // returns false if there's no controller
                this.getControllerLabel(ctlr.linkBank, 
                                        ctlr.linkChannel);
              if (!ctlr.label)
                post("\nmidi-faders.js::handleLinks: Error: Can't get " +
                     "controller linked from bank " + i + ", channel " + j + 
                     ".\nPoints to bank " + this.linkBank + ", channel " +
                     this.linkChannel);
            }
          }
        }
      }
    }
  }
  this.testScale = function(bank, channel) {
    var ctlr = this.getController(bank, channel, true, true);
    var s0 = ctlr.scale(0);
    var s64 = ctlr.scale(64);
    var s127 = ctlr.scale(127);
    post("\nscaled: 0 -> " + s0 + " descaled -> " + ctlr.descale(s0) +
         "\nscaled: 64 -> " + s64 + " descaled -> " + ctlr.descale(s64) +
         "\nscaled: 127 -> " + s127 + " descaled -> " + ctlr.descale(s127));
  }
  // 8.10.11: handle incoming pedal values to scale one of the controllers
  this.pedalMap = function(min, max, exponent, initialValue) {
    this.pedal= new controller(-1, min, max, exponent, initialValue,
                               "nowhere", "nothing", -1);
  }
  this.pedalBind = function (bank, channel) {
    // unbind the old if we set it
    var old = this.getController(this.pedalBank, this.pedalChannel, false,
                                 true);
    if (old)
      old.pedalBound = false;
    this.pedalBank = bank;
    this.pedalChannel = channel;
    var ctlr = this.getController(bank, channel, true, true);
    if (ctlr) {
      ctlr.pedalBound = true;
      this.pedalController = ctlr;
    }
  }
}
////////////////////////////////////////////////////////////////////////////////

// Constructor
function bank(bankNum) {
  this.bankNum = bankNum;
  // 17 because we won't bother starting at slot 0....
  this.controllers = new Array(17);
  this.print = function() {
    var i;
    for (i = 1; i < this.controllers.length; ++i) {
      if (this.controllers[i])
        this.controllers[i].print();
    }
  }
  // channel has just been called, now we can send out the message with scaled
  // value 
  this.storeController = function(controller, channel) {
    if (this.controllers[channel])
      post("midi-faders.js::bank::storeController: controller already stored " 
           + "at bank " + this.bankNum + ", controller " + channel +
           " but replacing anyway.");
    this.controllers[channel] = controller;
  }
  // call this when we change bank to dactivate all controllers so that next
  // time we're on this bank we have to get faders within tolerance of the
  // previous value 
  this.deactivate = function() {
    var i;
    for (i = 1; i < this.controllers.length; ++i) {
      if (this.controllers[i])
        this.controllers[i].deactivate();
    }
  }
  // if overwrite, set the label to just the controller number when there's
  // actually no controller in the bank.
  this.sendLabels = function(overwrite) {
    var i;
    for (i = 1; i < 17; ++i) {
      var ctlr = this.controllers[i];
      var label = ctlr ? i + ":" + ctlr.label : i;
      if (ctlr || overwrite)
        midiFadersSend("midi-faders-js-label" + i, 
                       MidiFadersLabel(label));
    }
  }
  this.displayCurrent = function() {
    var i;
    for (i = 1; i < 17; ++i) {
      var ctlr = this.controllers[i];
      if (ctlr)
        ctlr.displayCurrent();
    }
  }
}

////////////////////////////////////////////////////////////////////////////////

// Constructor
// the controller object holds data for a single controller
function controller(bank, min, max, exponent, initialValue, send, label, 
                    channel) {
  this.channel = channel; // 1-16
  this.bank = bank;
  // whether the previous fader value has been reached or not
  this.active = false;
  // 6.10.11 on or off state--not the same as active
  this.on = true;
  // data for scaling the input before outputting to the send
  this.scaledMin = min;
  this.scaledMax = max;
  this.scaledExpt = exponent;
  this.scaledRange = 0.0;
  // global send name
  this.send = send;
  // the label text, for the interface.  These will always be sent to the same
  // receiver on the interface patch (which if you use, you won't need to edit
  // in order for this to work properly).
  this.label = label;
  // the last 7bit MIDI value we saw
  this.last = 0;
  // 1.9.10 we can have a controller simply be a link to another by specifying
  // an exponent of -1
  this.isLink = exponent == -1;
  this.linkBank = -1;
  this.linkChannel = -1;
  // if initialValue is outside of range this won't have any effect as it's the
  // setValue call at the bottom of this constructor that uses this value to
  // set the current MIDI value (this.last)
  this.initialValue = initialValue;
  this.currentValue = initialValue;
  // does the incoming pedal value scale this controller?
  this.pedalBound = false;
  this.print = function() {
    var i;
    post("\nchannel: " + this.channel + ", bank: " + this.bank + ", last: "
         + this.last + ", scaledMin: " + this.scaledMin + ", scaledMax: " +
         this.scaledMax + ", scaledExpt: " + this.scaledExpt + 
         ",\nscaledRange: " + this.scaledRange +  ", send: " + this.send + 
         ", label: " + this.label + ", isLink: " + this.isLink + 
         ",\nlinkChannel: " + this.linkChannel + ", linkBank: " + this.linkBank
         + ", initialValue: " + this.initialValue + ", currentValue: " +
         this.currentValue);
  }
  this.deactivate = function() {
    this.active = false;
  }
  this.activate = function() {
    this.active = true;
  }
  this.setMin = function(min) {
    this.scaledMin = min;
    this.calcRange();
  }
  this.setMax = function(max) {
    this.scaledMax = max;
    this.calcRange();
  }
  this.setExpt = function(expt) {
    this.scaledExpt = expt;
  }
  this.calcRange = function() {
    this.scaledRange = this.scaledMax - this.scaledMin;
  }
  // scale a 7bit number to the desired range
  this.scale = function(val) {
    return this.scaledMin + 
    (this.scaledRange * Math.pow(val / 127.0, this.scaledExpt));
  }
  // take a number within the desired range and turn it back into the 7bit
  // number which would produce it if passed to scale().
  this.descale = function(val) {
    return Math.round(127.0 * 
                      Math.pow((val - this.scaledMin) / this.scaledRange,
                               (1.0 / this.scaledExpt)));
  }
  // check whether a scaled value is within min and max
  this.inRange = function(val, reallyWarn) {
    var ok = (val >= this.scaledMin && val <= this.scaledMax);
    if (!ok && MidiFadersConfig.warn && reallyWarn) {
      this.print();
      post("\nmidi-faders::controller::inRange: bank " + this.bank + " channel "
           + this.channel + " (" + this.label + ") value " + 
           val + " is not in range.");
    }
    return ok;
  }
  this.setValue = function(val) { // sets and outputs
    this.setValueAux(val, true);
  }
  // 18.10.11
  this.setValueNo = function(val) { // set but no output
    this.setValueAux(val, false);
  }
  this.setValueAux = function(val, output) {
    // this.print();
    // post("\nsetValueAux: val = " + val);
    if (this.inRange(val, true)) {
      var midi = this.descale(val);
      this.active = false;
      this.last = midi;
      if (output)
        this.output(val, 1.0, this.send);
    }
  }
  // 6.10.11: for toggling between 'off' (0) and 'on' (last value). Of course a
  // 0 might mean another value once scaled
  // MDE Mon Sep 23 13:52:24 2013 -- add arg to this function so we can set the
  // value instead of toggling 
  this.toggleOnOff = function(val) {
    // post("\ntoggleOnOff last " + this.last);
    if (this.on || !val) {
      this.output(this.scale(0), 1.0, this.send);
      this.on = false;
      // set text to blue
      midiFadersSend("midi-faders-js-number" + this.channel,
                     [ "textcolor", 0, 0, 1, 1 ]);
    }
    else {
      this.output(this.scale(this.last), 1.0, this.send);
      this.on = true;
      midiFadersSend("midi-faders-js-number" + this.channel,
                     [ "textcolor", 0, 0, 0, 1 ]);
    }
  }
  this.sendScaledVal = function(val) {
    if (!this.active && 
        val >= this.last - MidiFadersConfig.tolerance &&
        val <= this.last + MidiFadersConfig.tolerance)
      this.active = true;
    // if it's universal (i.e. bank == 0), just do it, otherise only if it's
    // within tolerance of last value
    if (!this.bank || this.active) {
      var scaled = this.scale(val);
      var pedalScaler = 1.0;
      // 8.10.11: if this controller is scaled by the pedal, do this here, but
      // don't store that value... 
      if (this.pedalBound) {
        pedalScaler = gMidiFadersBanks.pedalScaledVal;
      }
      this.output(scaled, pedalScaler, this.send);
      this.last = val;
      // set the display text to black
      midiFadersSend("midi-faders-js-number" + this.channel,
                     //             R  G  B  alpha
                     [ "textcolor", 0, 0, 0, 1 ]);
    }
    // else show target midi value in display so we know what we're aiming for
    else {
      midiFadersSend("midi-faders-js-number" + this.channel, this.last);
      // set the display text to red
      midiFadersSend("midi-faders-js-number" + this.channel,
                     [ "textcolor", 1, 0, 0, 1 ]);
    }
  }
  // the receiver is the unique process receive e.g. master-ampf, as defined in
  // the mapping file, but we'll also send to the interface channel by
  // e.g. midi-faders-js-number1 
  this.output = function(value, pedalScaler, receiver) {
    //    post("\noutput: value " + value + " pedalScaler " + pedalScaler +
    // " receiver " + receiver);
    var final = value * pedalScaler;
    midiFadersSend(this.send, final);
    outlet(0, final);
    this.currentValue = value; // i.e. don't store the pedal-scaled value
    // MDE Mon May  6 16:46:15 2013 -- only display value if this is the
    // current bank 
    if (!this.bank || this.bank == gMidiFadersBanks.currentBank)
      this.displayCurrent();
  }
  // 8.10.11: e.g. for pedal scaling
  this.rescale = function(scaler) {
    this.output(this.currentValue, scaler, this.send);
  }
  this.displayCurrent = function(value) {
    midiFadersSend("midi-faders-js-number" + this.channel, this.currentValue);
  }
  // if it is a link we only expect a mapping line like 
  // 2, 10 1 6 -1;
  // in this case, the 2 is still the bank for this controller, and 10 is the
  // channel, but instead of min and max, the third and fourth args are the
  // bank and channel of the controller this controller points to.
  if (this.isLink) {
    if (this.bank == 0) { // universal
      this.print();
      post("\nmidi-faders.js::controller: universal controllers (bank 0) " +
           "can't be links.");
    }
    this.linkBank = min;
    this.linkChannel = max;
    this.label = ""; // the label will be updated in handleLinks()
    this.send = "";
    this.min = 0;
    this.max = 0;
  }
  // set the range at instantiation
  this.calcRange();
  // the last 7-bit value we saw will be set here.
  if (!this.isLink && this.inRange(initialValue, false))
    this.setValue(initialValue);
}

////////////////////////////////////////////////////////////////////////////////
// global functions, callable from Max with a message e.g. [tolerance 5] 

// MDE Mon Sep 23 14:29:19 2013 
function setToggle(channel, value) {
  gMidiFadersBanks.setToggle(channel, value);
}

// how close the fader has to come to the previous stored value before it will
// start outputting again. 
function tolerance(val) {
  MidiFadersConfig.tolerance = val;
}
// set the value of a controller NB new value is within the scaled range, not
// the 7bit range--it will be converted before being stored.
function setValue(bank, channel, value) {
 //post("\nsetValue: bank " + bank + " channel " + channel + " value " + value);
  var controller = gMidiFadersBanks.getController(bank, channel, true, true);
  if (controller)
    controller.setValue(value);
}
function setValueNo(bank, channel, value) {
  var controller = gMidiFadersBanks.getController(bank, channel, true, true);
  if (controller)
    controller.setValueNo(value);
}
// change the minimum scaled value of a controller
function setMin(bank, channel, min) {
  var controller = gMidiFadersBanks.getController(bank, channel, true, true);
  if (controller) 
    controller.setMin(min);
}
// change the maximum scaled value of a controller
function setMax(bank, channel, max) {
  var controller = gMidiFadersBanks.getController(bank, channel, true, true);
  if (controller) 
    controller.setMax(max);
}
// change the exponent value of a controller
function setExpt(bank, channel, expt) {
  var controller = gMidiFadersBanks.getController(bank, channel, true, true);
  if (controller) 
    controller.setExpt(expt);
}
function changeBank(bankNum) {
  gMidiFadersBanks.changeBank(bankNum);
}
// NB due to the demands of max coll/list processing, channel/bank are in
// different order of the coll file.  This is called automatically when the
// coll file has been read so is not really for user interaction.
function store(channel, min, max, exponent, initialValue, send, label, bank) {
  // post("\n" + channel + " " + min + " " + max + " " + exponent + " " + 
  //       initialValue + " " + send + " " + label + " " + bank);
  // post("\nNew controller: bank " + bank + " channel " + channel);
  var ctlr = new controller(bank, min, max, exponent, initialValue, send,
                            label, channel);
  gMidiFadersBanks.storeController(ctlr, bank, channel);
}
function endStore() {
  gMidiFadersBanks.endStore();
}
function print() {
  gMidiFadersBanks.print();
}
function msg_int(i) {
  switch (inlet) {
  case 0:
    switchBank(i);
    break;
  case 1:
    gMidiFadersBanks.output(i);
    break;
  case 2:
    gMidiFadersBanks.channel(i);
    break;
  case 3:
    gMidiFadersBanks.toggleOnOff(i);
    break;
  case 4:
    gMidiFadersBanks.pedalIn(i);
    break;
  }
}
function testScale(bank, channel) {
  gMidiFadersBanks.testScale(bank, channel);
}

// 8.10.11
// handle incoming pedal values to scale one of the controllers
function pedalMap(min, max, exponent, initialValue)
{
  gMidiFadersBanks.pedalMap(min, max, exponent, initialValue);
  
}
function pedalBind(bank, channel)
{
  gMidiFadersBanks.pedalBind(bank, channel);
}
////////////////////////////////////////////////////////////////////////////////
// Local functions

// to set the text in a label (umenu) we need to send a list
// for some reason this doesn't work unless there's already an item in the umenu
MidiFadersLabel.local = 1;
function MidiFadersLabel(text) {
  return [ "setitem", 0, text];
}

midiFadersSend.local = 1;
function midiFadersSend(receiver, message) {
  MidiFadersGlobalMessenger.temp = message;
  MidiFadersGlobalMessenger.sendnamed(receiver, "temp");
}

// describeOutlet.local = 1;
function describeOutlet(num)
{
  switch (num) {
  case 0:
    assist("Scaled controller value (use for testing only).");
    break;
  }
}

// describeInlet.local = 1;
function describeInlet(num)
{
  switch (num) {
  case 0:
    assist("various messages: tolerance, store, changeBank, setValue...");
    break;
  case 1: 
    assist("MIDI controller value (always after channel!)");
    break;
  case 2:
    assist("channel (1-16)");
    break;
  case 3:
    assist("toggle on/off channels");
    break;
  }
}

// setoutletassist(-1, describeOutlet);
// setinletassist(-1, describeInlet);

////////////////////////////////////////////////////////////////////////////////
// EOF midi-fader.js
